@@ -268,7 +268,8 @@ var Search = {
         } else {
           // normal html builders
           requestUrl = DOCUMENTATION_OPTIONS.URL_ROOT + item[0] + DOCUMENTATION_OPTIONS.FILE_SUFFIX;
-          linkUrl = item[0] + DOCUMENTATION_OPTIONS.LINK_SUFFIX;
+          // If LINK_SUFFIX is undefined, default to ".html" - I actually don't see this being anything else.
+          linkUrl = item[0] + (DOCUMENTATION_OPTIONS.LINK_SUFFIX || ".html");
         }
         listItem.append($('<a/>').attr('href',
             linkUrl +
