@@ -67,7 +67,7 @@ def isMacOS10_12_5_OrLater():
     """ Return whether this system is macOS 10.12.5 or a later version. """
     try:
         return [int(x) for x in platform.mac_ver()[0].split(".")] >= [10, 12, 5]
-    except:
+    except ValueError:
         return False
 
 
@@ -93,7 +93,10 @@ def is_chromeos_garcon():
     # Linux filesystem. This uses Garcon, see for example
     # https://chromium.googlesource.com/chromiumos/platform2/+/master/vm_tools/garcon/#opening-urls
     # https://source.chromium.org/search?q=garcon-url-handler
-    return "garcon-url-handler" in webbrowser.get().name
+    try:
+        return "garcon-url-handler" in webbrowser.get().name
+    except AttributeError:
+        return False
 
 
 def run_fish_cmd(text):
@@ -191,30 +194,44 @@ def better_color(c1, c2):
 def parse_color(color_str):
     """A basic function to parse a color string, for example, 'red' '--bold'."""
     comps = color_str.split(" ")
-    color = "normal"
+    color = ""
     background_color = ""
     bold, underline, italics, dim, reverse = False, False, False, False, False
     for comp in comps:
         # Remove quotes
         comp = comp.strip("'\" ")
-        if comp == "--bold":
+        if comp == "--bold" or comp == "-o":
             bold = True
-        elif comp == "--underline":
+        elif comp == "--underline" or comp == "-u":
             underline = True
-        elif comp == "--italics":
+        elif comp == "--italics" or comp == "-i":
             italics = True
-        elif comp == "--dim":
+        elif comp == "--dim" or comp == "-d":
             dim = True
-        elif comp == "--reverse":
+        elif comp == "--reverse" or comp == "-r":
             reverse = True
-        elif comp.startswith("--background="):
+        elif comp.startswith("--background"):
             # Background color
-            background_color = better_color(
-                background_color, parse_one_color(comp[len("--background=") :])
-            )
+            c = comp[len("--background=") :]
+            parsed_c = parse_one_color(c)
+            # We prefer the unparsed version - if it says "brgreen", we use brgreen,
+            # instead of 00ff00
+            if better_color(background_color, parsed_c) == parsed_c:
+                background_color = c
+        elif comp.startswith("-b"):
+            # Background color in short.
+            skip = len("-b")
+            if comp[len("-b=")] in ["=", " "]:
+                skip += 1
+            c = comp[skip :]
+            parsed_c = parse_one_color(c)
+            if better_color(background_color, parsed_c) == parsed_c:
+                background_color = c
         else:
             # Regular color
-            color = better_color(color, parse_one_color(comp))
+            parsed_c = parse_one_color(comp)
+            if better_color(color, parsed_c) == parsed_c:
+                color = comp
 
     return {
         "color": color,
@@ -226,6 +243,26 @@ def parse_color(color_str):
         "reverse": reverse,
     }
 
+def unparse_color(col):
+    """A basic function to return the fish version of a color dict"""
+    if isinstance(col, str):
+        return col
+    ret = ""
+    if col["color"]:
+        ret += col["color"]
+    if col["bold"]:
+        ret += " --bold"
+    if col["underline"]:
+        ret += " --underline"
+    if col["italics"]:
+        ret += " --italics"
+    if col["dim"]:
+        ret += " --dim"
+    if col["reverse"]:
+        ret += " --reverse"
+    if col["background"]:
+        ret += " --background=" + col["background"]
+    return ret
 
 def parse_bool(val):
     val = val.lower()
@@ -239,10 +276,10 @@ def parse_bool(val):
 def html_color_for_ansi_color_index(val):
     arr = [
         "black",
-        "#AA0000",
-        "#00AA00",
+        "#FF0000",
+        "#00FF00",
         "#AA5500",
-        "#0000AA",
+        "#0000FF",
         "#AA00AA",
         "#00AAAA",
         "#AAAAAA",
@@ -508,26 +545,37 @@ g_special_escapes_dict = None
 def get_special_ansi_escapes():
     global g_special_escapes_dict
     if g_special_escapes_dict is None:
-        import curses
+        try:
+            import curses
 
-        g_special_escapes_dict = {}
-        curses.setupterm()
+            g_special_escapes_dict = {}
+            curses.setupterm()
 
-        # Helper function to get a value for a tparm
-        def get_tparm(key):
-            val = None
-            key = curses.tigetstr(key)
-            if key:
-                val = curses.tparm(key)
-            if val:
-                val = val.decode("utf-8")
-            return val
+            # Helper function to get a value for a tparm
+            def get_tparm(key):
+                val = None
+                key = curses.tigetstr(key)
+                if key:
+                    val = curses.tparm(key)
+                if val:
+                    val = val.decode("utf-8")
+                return val
 
-        # Just a few for now
-        g_special_escapes_dict["exit_attribute_mode"] = get_tparm("sgr0")
-        g_special_escapes_dict["bold"] = get_tparm("bold")
-        g_special_escapes_dict["underline"] = get_tparm("smul")
+            # Just a few for now
+            g_special_escapes_dict["exit_attribute_mode"] = get_tparm("sgr0")
+            g_special_escapes_dict["bold"] = get_tparm("bold")
+            g_special_escapes_dict["underline"] = get_tparm("smul")
+        except ImportError:
+            print("WARNING: The python curses module is missing.")
+            print("WARNING: Falling back to xterm-256color settings.")
+            print("WARNING: Rebuild python with curses headers!")
 
+            g_special_escapes_dict = {
+                'exit_attribute_mode': '\x1b(B\x1b[m',
+                'bold': '\x1b[1m',
+                'underline': '\x1b[4m'
+            }
+            
     return g_special_escapes_dict
 
 
@@ -865,7 +913,7 @@ class BindingParser:
         return readable_command + result
 
 
-class FishConfigTCPServer(SocketServer.TCPServer):
+class FishConfigTCPServer(SocketServer.ThreadingMixIn, SocketServer.TCPServer):
     """TCPServer that only accepts connections from localhost (IPv4/IPv6)."""
 
     WHITELIST = set(["::1", "::ffff:127.0.0.1", "127.0.0.1"])
@@ -880,7 +928,8 @@ class FishConfigHTTPRequestHandler(SimpleHTTPServer.SimpleHTTPRequestHandler):
     def write_to_wfile(self, txt):
         self.wfile.write(txt.encode("utf-8"))
 
-    def do_get_colors(self):
+    def do_get_colors(self, path=None):
+        """ Read the colors from a .theme file in path, or the current shell if no path has been given """
         # Looks for fish_color_*.
         # Returns an array of lists [color_name, color_description, color_value]
         result = []
@@ -902,7 +951,8 @@ class FishConfigHTTPRequestHandler(SimpleHTTPServer.SimpleHTTPRequestHandler):
                 "quote",
                 "redirection",
                 "valid_path",
-                "autosuggestion" "user",
+                "autosuggestion",
+                "user",
                 "host",
                 "cancel",
             ]
@@ -933,11 +983,34 @@ class FishConfigHTTPRequestHandler(SimpleHTTPServer.SimpleHTTPRequestHandler):
             "cancel": "The ^C cancel indicator",
         }
 
-        out, err = run_fish_cmd("set -L")
+        # If we don't have a path, we get the current theme.
+        if not path:
+            out, err = run_fish_cmd("set -L")
+        else:
+            with open(path) as f:
+                out = f.read()
+        extrainfo = {}
         for line in out.split("\n"):
+            # Ignore empty lines
+            if not line: continue
+            # Lines starting with "#" can contain metadata.
+            if line.startswith("#"):
+                if not ":" in line: continue
+                key, value = line.split(":", maxsplit=1)
+                key = key.strip("# '")
+                value = value.strip(" '\"")
+                # Only use keys we know
+                if not key in ("name", "preferred_background", "url"): continue
+                if key == "preferred_background":
+                    if value not in named_colors and not value.startswith("#"):
+                        value = "#" + value
+                extrainfo[key] = value
 
-            for match in re.finditer(r"^fish_color_(\S+) ?(.*)", line):
-                color_name, color_value = [x.strip() for x in match.group(1, 2)]
+
+            for match in re.finditer(r"^fish_(pager_)?color_(\S+) ?(.*)", line):
+                color_name, color_value = [x.strip() for x in match.group(2, 3)]
+                if match.group(1):
+                    color_name = "fish_pager_color_" + color_name
                 color_desc = descriptions.get(color_name, "")
                 data = {"name": color_name, "description": color_desc}
                 data.update(parse_color(color_value))
@@ -953,7 +1026,7 @@ class FishConfigHTTPRequestHandler(SimpleHTTPServer.SimpleHTTPRequestHandler):
             color_desc = descriptions.get(color_name, "")
             result.append([color_name, color_desc, parse_color("")])
 
-        return result
+        return result, extrainfo
 
     def do_get_functions(self):
         out, err = run_fish_cmd("functions")
@@ -1068,38 +1141,35 @@ class FishConfigHTTPRequestHandler(SimpleHTTPServer.SimpleHTTPRequestHandler):
         return result
 
     def do_get_color_for_variable(self, name):
-        # Return the color with the given name, or the empty string if there is
-        # none.
+        "Return the color with the given name, or the empty string if there is none."
         out, err = run_fish_cmd("echo -n $" + name)
         return out
 
-    def do_set_color_for_variable(
-        self, name, color, background_color, bold, underline, italics, dim, reverse
-    ):
+    def do_set_color_for_variable(self, name, color):
         "Sets a color for a fish color name, like 'autosuggestion'"
-        if not color:
+        if not name:
+            raise ValueError
+        if not color and not color == "":
             color = "normal"
-        varname = "fish_color_" + name
+        else:
+            color = unparse_color(color)
+        if not name.startswith("fish_pager_color_"):
+            varname = "fish_color_" + name
         # If the name already starts with "fish_", use it as the varname
         # This is needed for 'fish_pager_color' vars.
         if name.startswith("fish_"):
             varname = name
-        # TODO: Check if the varname is allowable.
+        # Check if the varname is allowable.
+        varname = varname.strip()
+        if not re.match("^[a-zA-Z0-9_]+$", varname):
+            print("Refusing to use variable name: '", varname, "'")
+            return
+        color = color.strip()
+        if not re.match("^[a-zA-Z0-9_= -]*$", color):
+            print("Refusing to use color value: ", color)
+            return
         command = "set -U " + varname
-        if color:
-            command += " " + color
-        if background_color:
-            command += " --background=" + background_color
-        if bold:
-            command += " --bold"
-        if underline:
-            command += " --underline"
-        if italics:
-            command += " --italics"
-        if dim:
-            command += " --dim"
-        if reverse:
-            command += " --reverse"
+        command += " " + color
 
         out, err = run_fish_cmd(command)
         return out
@@ -1118,19 +1188,25 @@ class FishConfigHTTPRequestHandler(SimpleHTTPServer.SimpleHTTPRequestHandler):
         return True
 
     def do_set_prompt_function(self, prompt_func):
-        cmd = prompt_func + "\n" + "funcsave fish_prompt"
+        cmd = "functions -e fish_right_prompt; " + prompt_func + "\n" + "funcsave fish_prompt && funcsave fish_right_prompt 2>/dev/null"
         out, err = run_fish_cmd(cmd)
         return len(err) == 0
 
-    def do_get_prompt(self, command_to_run, prompt_function_text, extras_dict):
+    def do_get_prompt(self, prompt_function_text, extras_dict):
         # Return the prompt output by the given command
-        prompt_demo_ansi, err = run_fish_cmd(command_to_run)
+        cmd = prompt_function_text + '\n builtin cd "' + initial_wd + '" \n false \n fish_prompt\n'
+        prompt_demo_ansi, err = run_fish_cmd(cmd)
         prompt_demo_html = ansi_to_html(prompt_demo_ansi)
-        prompt_demo_font_size = self.font_size_for_ansi_prompt(prompt_demo_ansi)
+        right_demo_ansi, err = run_fish_cmd(
+            "functions -e fish_right_prompt; " + prompt_function_text + '\n builtin cd "' + initial_wd + '" \n false \n functions -q fish_right_prompt && fish_right_prompt\n'
+        )
+        right_demo_html = ansi_to_html(right_demo_ansi)
+        prompt_demo_font_size = self.font_size_for_ansi_prompt(prompt_demo_ansi + right_demo_ansi)
         result = {
             "function": prompt_function_text,
             "demo": prompt_demo_html,
             "font_size": prompt_demo_font_size,
+            "right": right_demo_html,
         }
         if extras_dict:
             result.update(extras_dict)
@@ -1139,9 +1215,8 @@ class FishConfigHTTPRequestHandler(SimpleHTTPServer.SimpleHTTPRequestHandler):
     def do_get_current_prompt(self):
         # Return the current prompt. We run 'false' to demonstrate how the
         # prompt shows the command status (#1624).
-        prompt_func, err = run_fish_cmd("functions fish_prompt")
+        prompt_func, err = run_fish_cmd("functions fish_prompt; functions fish_right_prompt")
         result = self.do_get_prompt(
-            'builtin cd "' + initial_wd + '" ; false ; fish_prompt',
             prompt_func.strip(),
             {"name": "Current"},
         )
@@ -1151,8 +1226,7 @@ class FishConfigHTTPRequestHandler(SimpleHTTPServer.SimpleHTTPRequestHandler):
         # Return the prompt you get from the given text. Extras_dict is a
         # dictionary whose values get merged in. We run 'false' to demonstrate
         # how the prompt shows the command status (#1624)
-        cmd = text + '\n builtin cd "' + initial_wd + '" \n false \n fish_prompt\n'
-        return self.do_get_prompt(cmd, text.strip(), extras_dict)
+        return self.do_get_prompt(text.strip(), extras_dict)
 
     def parse_one_sample_prompt_hash(self, line, result_dict):
         # Allow us to skip whitespace, etc.
@@ -1198,7 +1272,7 @@ class FishConfigHTTPRequestHandler(SimpleHTTPServer.SimpleHTTPRequestHandler):
             return None
 
     def do_get_sample_prompts_list(self):
-        paths = glob.iglob("sample_prompts/*.fish")
+        paths = sorted(glob.iglob("sample_prompts/*.fish"))
         result = []
         try:
             pool = multiprocessing.pool.ThreadPool(processes=8)
@@ -1294,7 +1368,26 @@ class FishConfigHTTPRequestHandler(SimpleHTTPServer.SimpleHTTPRequestHandler):
         self.path = p
 
         if p == "/colors/":
-            output = self.do_get_colors()
+            # Construct our colorschemes.
+            # Add the current scheme first, then the default.
+            # The rest in alphabetical order.
+            curcolors, curinfo = self.do_get_colors()
+            defcolors, definfo = self.do_get_colors("themes/fish default.theme")
+            curinfo.update({ "theme": "Current", "colors": curcolors})
+            definfo.update({ "theme": "fish default", "colors": defcolors})
+            output = [curinfo, definfo]
+
+            confighome = os.environ["XDG_CONFIG_HOME"] if "XDG_CONFIG_HOME" in os.environ else os.path.expanduser("~")
+            paths = list(glob.iglob(os.path.join(confighome, "fish", "themes/*.theme")))
+            paths.extend(list(glob.iglob("themes/*.theme")))
+            paths.sort(key=str.casefold)
+
+            for p in paths:
+                theme = os.path.splitext(os.path.basename(p))[0]
+                if any(theme == d["theme"] for d in output): continue
+                out, outinfo = self.do_get_colors(p)
+                outinfo.update({ "theme": theme, "colors": out })
+                output.append(outinfo)
         elif p == "/functions/":
             output = self.do_get_functions()
         elif p == "/variables/":
@@ -1363,29 +1456,14 @@ class FishConfigHTTPRequestHandler(SimpleHTTPServer.SimpleHTTPRequestHandler):
             postvars = {}
 
         if p == "/set_color/":
-            what = postvars.get("what")
-            color = postvars.get("color")
-            background_color = postvars.get("background_color")
-            bold = postvars.get("bold")
-            italics = postvars.get("italics")
-            reverse = postvars.get("reverse")
-            dim = postvars.get("dim")
-            underline = postvars.get("underline")
+            print("# Colorscheme: " + postvars.get("theme"))
+            for item in postvars.get("colors"):
+                what = item.get("what")
+                color = item.get("color")
+
+                if what:
+                    output = self.do_set_color_for_variable(what, color)
 
-            if what:
-                # Not sure why we get lists here?
-                output = self.do_set_color_for_variable(
-                    what[0],
-                    color[0],
-                    background_color[0],
-                    parse_bool(bold[0]),
-                    parse_bool(underline[0]),
-                    parse_bool(italics[0]),
-                    parse_bool(dim[0]),
-                    parse_bool(reverse[0]),
-                )
-            else:
-                output = "Bad request"
         elif p == "/get_function/":
             what = postvars.get("what")
             output = [self.do_get_function(what[0])]
@@ -1430,9 +1508,9 @@ class FishConfigHTTPRequestHandler(SimpleHTTPServer.SimpleHTTPRequestHandler):
         pass
 
     def log_error(self, format, *args):
-        if format == "code %d, message %s":
+        if format == "code %d, message %s" and hasattr(self, 'path'):
             # This appears to be a send_error() message
-            # We want to include the path
+            # We want to include the path (if we have one)
             (code, msg) = args
             format = "code %d, message %s, path %s"
             args = (code, msg, self.path)
@@ -1522,7 +1600,7 @@ if len(sys.argv) > 1:
         "abbreviations",
     ]:
         if tab.startswith(sys.argv[1]):
-            initial_tab = "#" + tab
+            initial_tab = "#!/" + tab
             break
 
 url = "http://localhost:%d/%s/%s" % (PORT, authkey, initial_tab)
@@ -1570,6 +1648,11 @@ def runThing():
 thread = threading.Thread(target=runThing)
 thread.start()
 
+# Safari will open sockets and not write to them, causing potential hangs
+# on shutdown.
+httpd.block_on_close = False
+httpd.daemon_threads = True
+
 # Select on stdin and httpd
 stdin_no = sys.stdin.fileno()
 try:
